# coding: utf-8
from datetime import datetime
from unittest import TestCase

from mock import (
    ANY,
    MagicMock,
    call,
    patch,
)

from fireant.slicer.operations import Totals
from fireant.slicer.transformers import (
    PandasRowIndexTransformer,
    TransformationException,
)
from fireant.tests import mock_dataframes as mock_df


class PandasRowIndexTransformerTests(TestCase):
    pd_tx = PandasRowIndexTransformer()

    def test_no_dims_multi_metric(self):
        # Tests transformation of a single metric with a single continuous dimension
        result = self.pd_tx.transform(mock_df.no_dims_multi_metric_df, mock_df.no_dims_multi_metric_schema)

        self.assertListEqual([None], list(result.index.names))
        self.assertListEqual([0], list(result.index))

        self.assertListEqual(['One', 'Two', 'Three', 'Four', 'Five', 'Six', 'Seven', 'Eight'], list(result.columns))
        self.assertListEqual([a for a in range(8)], result.values[0].tolist())

    def test_cont_dim_single_metric(self):
        # Tests transformation of a single metric with a single continuous dimension
        result = self.pd_tx.transform(mock_df.cont_dim_single_metric_df, mock_df.cont_dim_single_metric_schema)

        self.assertListEqual(['Cont'], list(result.index.names))
        self.assertListEqual([a for a in range(8)], list(result.index))

        self.assertListEqual(['One'], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result['One']))

    def test_cont_dim_multi_metric(self):
        # Tests transformation of two metrics with a single continuous dimension
        result = self.pd_tx.transform(mock_df.cont_dim_multi_metric_df, mock_df.cont_dim_multi_metric_schema)

        self.assertListEqual(['Cont'], list(result.index.names))
        self.assertListEqual([a for a in range(8)], list(result.index))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result['One']))
        self.assertListEqual([2 * a for a in range(8)], list(result['Two']))

    def test_time_series_date(self):
        # Tests transformation of a single-metric, single-dimension result
        result = self.pd_tx.transform(mock_df.time_dim_single_metric_df, mock_df.time_dim_single_metric_schema)

        self.assertListEqual(['Date'], list(result.index.names))
        self.assertListEqual([datetime(2000, 1, i) for i in range(1, 9)], list(result.index))

        self.assertListEqual(['One'], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result['One']))

    def test_time_series_date_with_ref(self):
        # Tests transformation of a single-metric, single-dimension result using a WoW reference
        result = self.pd_tx.transform(mock_df.time_dim_single_metric_ref_df, mock_df.time_dim_single_metric_ref_schema)

        self.assertListEqual(['Date'], list(result.index.names))
        self.assertListEqual([datetime(2000, 1, i) for i in range(1, 9)], list(result.index))

        self.assertListEqual([('One', None), ('One', 'WoW')], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result[('One', None)]))
        self.assertListEqual([2 * a for a in range(8)], list(result[('One', 'WoW')]))

    def test_uni_dim_single_metric(self):
        # Tests transformation of a metric with a unique dimension with one key and display
        result = self.pd_tx.transform(mock_df.uni_dim_single_metric_df, mock_df.uni_dim_single_metric_schema)

        self.assertListEqual(['Uni'], list(result.index.names))
        self.assertListEqual(['One'], list(result.columns))

    def test_uni_dim_multi_metric(self):
        # Tests transformation of a metric with a unique dimension with one key and display
        result = self.pd_tx.transform(mock_df.uni_dim_multi_metric_df, mock_df.uni_dim_multi_metric_schema)

        self.assertListEqual(['Uni'], list(result.index.names))
        self.assertListEqual(['Aa', 'Bb', 'Cc'], list(result.index))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(3)], list(result['One']))
        self.assertListEqual([2 * a for a in range(3)], list(result['Two']))

    def test_cat_cat_dim_single_metric(self):
        # Tests transformation of a single metric with two categorical dimensions
        result = self.pd_tx.transform(mock_df.cat_cat_dims_single_metric_df, mock_df.cat_cat_dims_single_metric_schema)

        self.assertListEqual(['Cat1', 'Cat2'], list(result.index.names))
        self.assertListEqual(['A', 'B'], list(result.index.levels[0]))
        self.assertListEqual(['Y', 'Z'], list(result.index.levels[1]))

        self.assertListEqual(['One'], list(result.columns))
        self.assertListEqual([a for a in range(4)], list(result['One']))

    def test_cat_cat_dim_multi_metric(self):
        # Tests transformation of two metrics with two categorical dimensions
        result = self.pd_tx.transform(mock_df.cat_cat_dims_multi_metric_df, mock_df.cat_cat_dims_multi_metric_schema)

        self.assertListEqual(['Cat1', 'Cat2'], list(result.index.names))
        self.assertListEqual(['A', 'B'], list(result.index.levels[0]))
        self.assertListEqual(['Y', 'Z'], list(result.index.levels[1]))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(4)], list(result['One']))
        self.assertListEqual([2 * a for a in range(4)], list(result['Two']))

    def test_rollup_cont_cat_cat_dim_multi_metric(self):
        # Tests transformation of two metrics with two categorical dimensions
        df = mock_df.rollup_cont_cat_cat_dims_multi_metric_df

        result = self.pd_tx.transform(df, mock_df.rollup_cont_cat_cat_dims_multi_metric_schema)

        self.assertListEqual(['Cont', 'Cat1', 'Cat2'], list(result.index.names))
        self.assertListEqual([0, 1, 2, 3, 4, 5, 6, 7], list(result.index.levels[0]))
        self.assertListEqual([Totals.label, 'A', 'B'], list(result.index.levels[1]))
        self.assertListEqual([Totals.label, 'Y', 'Z'], list(result.index.levels[2]))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([12, 1, 0, 1, 5, 2, 3,
                              44, 9, 4, 5, 13, 6, 7,
                              76, 17, 8, 9, 21, 10, 11,
                              108, 25, 12, 13, 29, 14, 15,
                              140, 33, 16, 17, 37, 18, 19,
                              172, 41, 20, 21, 45, 22, 23,
                              204, 49, 24, 25, 53, 26, 27,
                              236, 57, 28, 29, 61, 30, 31], list(result['One']))
        self.assertListEqual([24, 2, 0, 2, 10, 4, 6,
                              88, 18, 8, 10, 26, 12, 14,
                              152, 34, 16, 18, 42, 20, 22,
                              216, 50, 24, 26, 58, 28, 30,
                              280, 66, 32, 34, 74, 36, 38,
                              344, 82, 40, 42, 90, 44, 46,
                              408, 98, 48, 50, 106, 52, 54,
                              472, 114, 56, 58, 122, 60, 62], list(result['Two']))

    def test_empty_dataframe_is_handled(self):
        df = mock_df.cat_cat_dims_single_metric_empty_df

        result = self.pd_tx.transform(df, mock_df.cat_cat_dims_single_metric_schema)

        self.assertTrue(result.empty)
        self.assertEqual(result.index.names, ['Cat1', 'Cat2'])
        self.assertEqual(list(result.index.levels[0]), [])
        self.assertEqual(list(result.index.levels[1]), [])
        self.assertEqual(list(result.columns), ['One'])


class PandasColumnIndexTransformerTests(TestCase):
    pd_tx = PandasRowIndexTransformer()

    def test_no_dims_multi_metric(self):
        # Tests transformation of a single metric with a single continuous dimension
        result = self.pd_tx.transform(mock_df.no_dims_multi_metric_df, mock_df.no_dims_multi_metric_schema)

        self.assertListEqual([None], list(result.index.names))
        self.assertListEqual([0], list(result.index))

        self.assertListEqual(['One', 'Two', 'Three', 'Four', 'Five', 'Six', 'Seven', 'Eight'], list(result.columns))
        self.assertListEqual([a for a in range(8)], result.values[0].tolist())

    def test_cont_dim_single_metric(self):
        # Tests transformation of a single metric with a single continuous dimension
        result = self.pd_tx.transform(mock_df.cont_dim_single_metric_df, mock_df.cont_dim_single_metric_schema)

        self.assertListEqual(['Cont'], list(result.index.names))
        self.assertListEqual([a for a in range(8)], list(result.index))

        self.assertListEqual(['One'], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result['One']))

    def test_cont_dim_multi_metric(self):
        # Tests transformation of two metrics with a single continuous dimension
        result = self.pd_tx.transform(mock_df.cont_dim_multi_metric_df, mock_df.cont_dim_multi_metric_schema)

        self.assertListEqual(['Cont'], list(result.index.names))
        self.assertListEqual([a for a in range(8)], list(result.index))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result['One']))
        self.assertListEqual([2 * a for a in range(8)], list(result['Two']))

    def test_time_series_date_to_millis(self):
        # Tests transformation of a single-metric, single-dimension result
        result = self.pd_tx.transform(mock_df.time_dim_single_metric_df, mock_df.time_dim_single_metric_schema)

        self.assertListEqual(['Date'], list(result.index.names))
        self.assertListEqual([datetime(2000, 1, i) for i in range(1, 9)], list(result.index))

        self.assertListEqual(['One'], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result['One']))

    def test_time_series_date_with_ref(self):
        # Tests transformation of a single-metric, single-dimension result using a WoW reference
        result = self.pd_tx.transform(mock_df.time_dim_single_metric_ref_df, mock_df.time_dim_single_metric_ref_schema)

        self.assertListEqual(['Date'], list(result.index.names))
        self.assertListEqual([datetime(2000, 1, i) for i in range(1, 9)], list(result.index))

        self.assertListEqual([('One', None), ('One', 'WoW')], list(result.columns))
        self.assertListEqual([a for a in range(8)], list(result[('One', None)]))
        self.assertListEqual([2 * a for a in range(8)], list(result[('One', 'WoW')]))

    def test_cont_cat_dim_single_metric(self):
        # Tests transformation of a single metric with a continuous and a categorical dimension
        result = self.pd_tx.transform(mock_df.cont_cat_dims_single_metric_df,
                                      mock_df.cont_cat_dims_single_metric_schema)

        self.assertListEqual(['Cont', 'Cat1'], list(result.index.names))
        self.assertListEqual([(a, b)
                              for a in range(8)
                              for b in ['A', 'B']], list(result.index))

        self.assertListEqual(['One'], list(result.columns))
        self.assertListEqual([a for a in range(16)], list(result['One']))

    def test_cont_cat_dim_multi_metric(self):
        # Tests transformation of two metrics with a continuous and a categorical dimension
        result = self.pd_tx.transform(mock_df.cont_cat_dims_multi_metric_df, mock_df.cont_cat_dims_multi_metric_schema)

        self.assertListEqual(['Cont', 'Cat1'], list(result.index.names))
        self.assertListEqual([(a, b)
                              for a in range(8)
                              for b in ['A', 'B']], list(result.index))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(16)], list(result['One']))
        self.assertListEqual([2 * a for a in range(16)], list(result['Two']))

    def test_cont_cat_cat_dim_multi_metric(self):
        # Tests transformation of two metrics with a continuous and two categorical dimensions
        result = self.pd_tx.transform(mock_df.cont_cat_cat_dims_multi_metric_df,
                                      mock_df.cont_cat_cat_dims_multi_metric_schema)

        self.assertListEqual(['Cont', 'Cat1', 'Cat2'], list(result.index.names))
        self.assertListEqual([(a, b, c)
                              for a in range(8)
                              for b in ['A', 'B']
                              for c in ['Y', 'Z']], list(result.index))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(32)], list(result['One']))
        self.assertListEqual([2 * a for a in range(32)], list(result['Two']))

    def test_cont_cat_uni_dim_multi_metric(self):
        # Tests transformation of two metrics with a continuous and two categorical dimensions
        result = self.pd_tx.transform(mock_df.cont_cat_uni_dims_multi_metric_df,
                                      mock_df.cont_cat_uni_dims_multi_metric_schema)

        self.assertListEqual(['Cont', 'Cat1', 'Uni'], list(result.index.names))
        self.assertListEqual([(a, b, c)
                              for a in range(8)
                              for b in ['A', 'B']
                              for c in ['Aa', 'Bb', 'Cc']], list(result.index))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([a for a in range(48)], list(result['One']))
        self.assertListEqual([a for a in range(100, 148)], list(result['Two']))

    def test_rollup_cont_cat_cat_dims_multi_metric_df(self):
        # Tests transformation of two metrics with a continuous and two categorical dimensions
        result = self.pd_tx.transform(mock_df.rollup_cont_cat_cat_dims_multi_metric_df,
                                      mock_df.rollup_cont_cat_cat_dims_multi_metric_schema)

        self.assertListEqual(['Cont', 'Cat1', 'Cat2'], list(result.index.names))
        self.assertListEqual([0, 1, 2, 3, 4, 5, 6, 7], list(result.index.levels[0]))
        self.assertListEqual([Totals.label, 'A', 'B'], list(result.index.levels[1]))
        self.assertListEqual([Totals.label, 'Y', 'Z'], list(result.index.levels[2]))

        self.assertListEqual(['One', 'Two'], list(result.columns))
        self.assertListEqual([12, 1, 0, 1, 5, 2, 3,
                              44, 9, 4, 5, 13, 6, 7,
                              76, 17, 8, 9, 21, 10, 11,
                              108, 25, 12, 13, 29, 14, 15,
                              140, 33, 16, 17, 37, 18, 19,
                              172, 41, 20, 21, 45, 22, 23,
                              204, 49, 24, 25, 53, 26, 27,
                              236, 57, 28, 29, 61, 30, 31], list(result['One']))
        self.assertListEqual([24, 2, 0, 2, 10, 4, 6,
                              88, 18, 8, 10, 26, 12, 14,
                              152, 34, 16, 18, 42, 20, 22,
                              216, 50, 24, 26, 58, 28, 30,
                              280, 66, 32, 34, 74, 36, 38,
                              344, 82, 40, 42, 90, 44, 46,
                              408, 98, 48, 50, 106, 52, 54,
                              472, 114, 56, 58, 122, 60, 62], list(result['Two']))


class MatplotlibLineChartTransformerTests(TestCase):
    mock_matplotlib = MagicMock(name='mock_matplotlib')
    patcher = patch.dict('sys.modules', {
        'matplotlib': mock_matplotlib,
        'matplotlib.pyplot': mock_matplotlib.pyplot,
    })

    def setUp(self):
        self.patcher.start()

        mock_axes = MagicMock()
        mock_axes.__iter__.return_value = iter([0, 1, 2, 3])
        self.mock_matplotlib.pyplot.subplots.return_value = MagicMock(), mock_axes

    def tearDown(self):
        self.patcher.stop()

    from fireant.slicer.transformers import MatplotlibLineChartTransformer
    plt_tx = MatplotlibLineChartTransformer()

    def _assert_matplotlib_calls(self, mock_plot, metrics):
        if 1 == len(metrics):
            calls = [call(figsize=(14, 5)),
                     call().legend(loc='center left', bbox_to_anchor=(1, 0.5)),
                     call().legend().set_title(metrics[0])]
        else:
            calls = [c
                     for metric in metrics
                     for c in [call(ax=ANY),
                               call().legend(loc='center left', bbox_to_anchor=(1, 0.5)),
                               call().legend().set_title(metric)]]

        mock_plot.line.assert_has_calls(calls)

    @patch('pandas.DataFrame.plot')
    def test_series_single_metric(self, mock_plot):
        # Tests transformation of a single-metric, single-dimension result
        result = self.plt_tx.transform(mock_df.cont_dim_single_metric_df, mock_df.cont_dim_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One'])

    @patch('pandas.Series.plot')
    def test_series_multi_metric(self, mock_plot):
        # Tests transformation of a multi-metric, single-dimension result
        result = self.plt_tx.transform(mock_df.cont_dim_multi_metric_df, mock_df.cont_dim_multi_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One', 'Two'])

    @patch('pandas.DataFrame.plot')
    def test_time_series_date_to_millis(self, mock_plot):
        # Tests transformation of a single-metric, single-dimension result
        result = self.plt_tx.transform(mock_df.time_dim_single_metric_df, mock_df.time_dim_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One'])

    @patch('pandas.DataFrame.plot')
    def test_time_series_date_with_ref(self, mock_plot):
        # Tests transformation of a single-metric, single-dimension result using a WoW reference
        result = self.plt_tx.transform(mock_df.time_dim_single_metric_ref_df, mock_df.time_dim_single_metric_ref_schema)

        self._assert_matplotlib_calls(mock_plot, ['One'])

    @patch('pandas.DataFrame.plot')
    def test_cont_uni_dim_single_metric(self, mock_plot):
        # Tests transformation of a metric and a unique dimension
        result = self.plt_tx.transform(mock_df.cont_uni_dims_single_metric_df,
                                       mock_df.cont_uni_dims_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One'])

    @patch('pandas.DataFrame.plot')
    def test_cont_uni_dim_multi_metric(self, mock_plot):
        # Tests transformation of two metrics and a unique dimension
        result = self.plt_tx.transform(mock_df.cont_uni_dims_multi_metric_df, mock_df.cont_uni_dims_multi_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One', 'Two'])

    @patch('pandas.DataFrame.plot')
    def test_double_dimension_single_metric(self, mock_plot):
        # Tests transformation of a single-metric, double-dimension result
        result = self.plt_tx.transform(mock_df.cont_cat_dims_single_metric_df,
                                       mock_df.cont_cat_dims_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One'])

    @patch('pandas.DataFrame.plot')
    def test_double_dimension_multi_metric(self, mock_plot):
        # Tests transformation of a multi-metric, double-dimension result
        result = self.plt_tx.transform(mock_df.cont_cat_dims_multi_metric_df, mock_df.cont_cat_dims_multi_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One', 'Two'])

    @patch('pandas.DataFrame.plot')
    def test_triple_dimension_multi_metric(self, mock_plot):
        # Tests transformation of a multi-metric, double-dimension result
        df = mock_df.cont_cat_cat_dims_multi_metric_df

        result = self.plt_tx.transform(df, mock_df.cont_cat_cat_dims_multi_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One', 'Two'])

    @patch('pandas.DataFrame.plot')
    def test_rollup_triple_dimension_multi_metric(self, mock_plot):
        # Tests transformation of a multi-metric, double-dimension result
        df = mock_df.rollup_cont_cat_cat_dims_multi_metric_df

        result = self.plt_tx.transform(df, mock_df.rollup_cont_cat_cat_dims_multi_metric_schema)

        self._assert_matplotlib_calls(mock_plot, ['One', 'Two'])

    def test_require_at_least_one_dimension(self):
        df = mock_df.no_dims_multi_metric_df

        with self.assertRaises(TransformationException):
            self.plt_tx.transform(df, mock_df.no_dims_multi_metric_schema)


class MatplotlibBarChartTransformerTests(TestCase):
    mock_matplotlib = MagicMock(name='mock_matplotlib')
    patcher = patch.dict('sys.modules', {
        'matplotlib': mock_matplotlib,
        'matplotlib.pyplot': mock_matplotlib.pyplot,
    })

    def setUp(self):
        self.patcher.start()

        mock_axes = MagicMock()
        mock_axes.__iter__.return_value = iter([0, 1, 2, 3])
        self.mock_matplotlib.pyplot.subplots.return_value = MagicMock(), mock_axes

    def tearDown(self):
        self.patcher.stop()

    from fireant.slicer.transformers import MatplotlibBarChartTransformer
    plt_tx = MatplotlibBarChartTransformer()

    def _assert_matplotlib_calls(self, mock_plot):
        mock_plot.bar.assert_has_calls([call(figsize=(14, 5)),
                                        call().legend(loc='center left', bbox_to_anchor=(1, 0.5))])

    @patch('pandas.DataFrame.plot')
    def test_series_single_metric(self, mock_plot):
        # Tests transformation of a single-metric, single-dimension result
        result = self.plt_tx.transform(mock_df.cont_dim_single_metric_df, mock_df.cont_dim_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot)

    @patch('pandas.DataFrame.plot')
    def test_series_multi_metric(self, mock_plot):
        # Tests transformation of a multi-metric, single-dimension result
        result = self.plt_tx.transform(mock_df.cont_dim_multi_metric_df, mock_df.cont_dim_multi_metric_schema)

        self._assert_matplotlib_calls(mock_plot)

    @patch('pandas.DataFrame.plot')
    def test_time_series_date_to_millis(self, mock_plot):
        # Tests transformation of a single-metric, single-dimension result
        result = self.plt_tx.transform(mock_df.time_dim_single_metric_df, mock_df.time_dim_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot)

    @patch('pandas.DataFrame.plot')
    def test_time_series_date_with_ref(self, mock_plot):
        # Tests transformation of a single-metric, single-dimension result using a WoW reference
        result = self.plt_tx.transform(mock_df.time_dim_single_metric_ref_df, mock_df.time_dim_single_metric_ref_schema)

        self._assert_matplotlib_calls(mock_plot)

    @patch('pandas.DataFrame.plot')
    def test_cont_uni_dim_single_metric(self, mock_plot):
        # Tests transformation of a metric and a unique dimension
        result = self.plt_tx.transform(mock_df.cont_uni_dims_single_metric_df,
                                       mock_df.cont_uni_dims_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot)

    @patch('pandas.DataFrame.plot')
    def test_double_dimension_single_metric(self, mock_plot):
        # Tests transformation of a single-metric, double-dimension result
        result = self.plt_tx.transform(mock_df.cont_cat_dims_single_metric_df,
                                       mock_df.cont_cat_dims_single_metric_schema)

        self._assert_matplotlib_calls(mock_plot)
